#!/usr/bin/env python

"""
Web interface data abstractions.

Copyright (C) 2014, 2015 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from datetime import datetime, timedelta
from imiptools.dates import format_datetime, get_datetime, get_start_of_day, to_date
from imiptools.period import Period

class PeriodError(Exception):
    pass

class EventPeriod(Period):

    """
    A simple period plus attribute details, compatible with RecurringPeriod, and
    intended to represent information obtained from an iCalendar resource.
    """

    def __init__(self, start, end, start_attr=None, end_attr=None, form_start=None, form_end=None, origin=None):
        Period.__init__(self, start, end)
        self.start_attr = start_attr
        self.end_attr = end_attr
        self.form_start = form_start
        self.form_end = form_end
        self.origin = origin

    def as_tuple(self):
        return self.start, self.end, self.start_attr, self.end_attr, self.form_start, self.form_end, self.origin

    def __repr__(self):
        return "EventPeriod(%r, %r, %r, %r, %r, %r, %r)" % self.as_tuple()

    def as_event_period(self):
        return self

    # Period data methods.

    def get_start(self):
        return self.start

    def get_end(self):
        return end_date_from_calendar(self.end)

    # Form data compatibility methods.

    def get_form_start(self):
        if not self.form_start:
            self.form_start = self.get_form_date(self.get_start(), self.start_attr)
        return self.form_start

    def get_form_end(self):
        if not self.form_end:
            self.form_end = self.get_form_date(self.get_end(), self.end_attr)
        return self.form_end

    def as_form_period(self):
        return FormPeriod(
            self.get_form_start(),
            self.get_form_end(),
            isinstance(self.end, datetime) or self.get_start() != self.get_end(),
            isinstance(self.start, datetime) or isinstance(self.end, datetime)
            )

    def get_form_date(self, dt, attr=None):
        return FormDate(
            format_datetime(to_date(dt)),
            isinstance(dt, datetime) and str(dt.hour) or None,
            isinstance(dt, datetime) and str(dt.minute) or None,
            isinstance(dt, datetime) and str(dt.second) or None,
            attr and attr.get("TZID") or None,
            dt, attr
            )

class FormPeriod:

    "A period whose information originates from a form."

    def __init__(self, start, end, end_enabled=True, times_enabled=True, origin=None):
        self.start = start
        self.end = end
        self.end_enabled = end_enabled
        self.times_enabled = times_enabled
        self.origin = origin

    def as_tuple(self):
        return self.start, self.end, self.end_enabled, self.times_enabled, self.origin

    def __repr__(self):
        return "FormPeriod(%r, %r, %r, %r, %r)" % self.as_tuple()

    def _get_start(self):
        t = self.start.as_datetime_item(self.times_enabled)
        if t:
            return t
        else:
            return None

    def _get_end(self, adjust=False):

        # Handle specified end datetimes.

        if self.end_enabled:
            t = self.end.as_datetime_item(self.times_enabled)
            if t:
                dtend, dtend_attr = t
                dtend = adjust and end_date_to_calendar(dtend) or dtend
            else:
                return None

        # Otherwise, treat the end date as the start date. Datetimes are
        # handled by making the event occupy the rest of the day.

        else:
            t = self._get_start()
            if t:
                dtstart, dtstart_attr = t
                dtend = dtstart + timedelta(1)
                dtend_attr = dtstart_attr

                if isinstance(dtstart, datetime):
                    dtend = get_start_of_day(dtend, dtend_attr["TZID"])
            else:
                return None

        return dtend, dtend_attr

    def as_event_period(self, index=None):
        t = self._get_start()
        if t:
            dtstart, dtstart_attr = t
        else:
            raise PeriodError(*[index is not None and ("dtstart", index) or "dtstart"])

        t = self._get_end(adjust=True)
        if t:
            dtend, dtend_attr = t
        else:
            raise PeriodError(*[index is not None and ("dtend", index) or "dtend"])

        if dtstart > dtend:
            raise PeriodError(*[
                index is not None and ("dtstart", index) or "dtstart",
                index is not None and ("dtend", index) or "dtend"
                ])

        return EventPeriod(dtstart, dtend, dtstart_attr, dtend_attr, self.start, self.end)

    # Period data methods.

    def get_start(self):
        t = self._get_start()
        if t:
            dtstart, dtstart_attr = t
            return dtstart
        else:
            return None

    def get_end(self):
        t = self._get_end()
        if t:
            dtend, dtend_attr = t
            return dtend
        else:
            return None

    # Form data methods.

    def get_form_start(self):
        return self.start

    def get_form_end(self):
        return self.end

    def as_form_period(self):
        return self

class FormDate:

    "Date information originating from form information."

    def __init__(self, date=None, hour=None, minute=None, second=None, tzid=None, dt=None, attr=None):
        self.date = date
        self.hour = hour
        self.minute = minute
        self.second = second
        self.tzid = tzid
        self.dt = dt
        self.attr = attr

    def as_tuple(self):
        return self.date, self.hour, self.minute, self.second, self.tzid, self.dt, self.attr

    def __repr__(self):
        return "FormDate(%r, %r, %r, %r, %r, %r, %r)" % self.as_tuple()

    def get_component(self, value):
        return (value or "").rjust(2, "0")[:2]

    def get_hour(self):
        return self.get_component(self.hour)

    def get_minute(self):
        return self.get_component(self.minute)

    def get_second(self):
        return self.get_component(self.second)

    def get_date_string(self):
        return self.date or ""

    def get_datetime_string(self):
        if not self.date:
            return ""

        hour = self.hour; minute = self.minute; second = self.second

        if hour or minute or second:
            time = "T%s%s%s" % tuple(map(self.get_component, (hour, minute, second)))
        else:
            time = ""
            
        return "%s%s" % (self.date, time)

    def get_tzid(self):
        return self.tzid

    def as_datetime_item(self, with_time=True):

        """
        Return a (datetime, attr) tuple for the datetime information provided by
        this object, or None if the fields cannot be used to construct a
        datetime object.
        """

        # Return any original datetime details.

        if self.dt:
            return self.dt, self.attr

        # Otherwise, construct a datetime and attributes.

        if not self.date:
            return None
        elif with_time:
            attr = {"TZID" : self.get_tzid(), "VALUE" : "DATE-TIME"}
            dt = get_datetime(self.get_datetime_string(), attr)
        else:
            dt = None

        # Interpret incomplete datetimes as dates.

        if not dt:
            attr = {"VALUE" : "DATE"}
            dt = get_datetime(self.get_date_string(), attr)

        if dt:
            return dt, attr

        return None

def end_date_to_calendar(dt):

    """
    Change end dates to refer to the iCalendar "next day" dates, not the actual
    dates.
    """

    if not isinstance(dt, datetime):
        return dt + timedelta(1)
    else:
        return dt

def end_date_from_calendar(dt):

    """
    Change end dates to refer to the actual dates, not the iCalendar "next day"
    dates.
    """

    if not isinstance(dt, datetime):
        return dt - timedelta(1)
    else:
        return dt

def event_period_from_period(period):
    if isinstance(period, EventPeriod):
        return period
    elif isinstance(period, FormPeriod):
        return period.as_event_period()
    else:
        return EventPeriod(period.start, period.end, period.start_attr, period.end_attr, origin=period.origin)

def form_period_from_period(period):
    if isinstance(period, EventPeriod):
        return period.as_form_period()
    elif isinstance(period, FormPeriod):
        return period
    else:
        return event_period_from_period(period).as_form_period()

# vim: tabstop=4 expandtab shiftwidth=4
